import logging

l = logging.getLogger("archr.analyzers.datascout")

from ..errors import ArchrError
from . import Analyzer

# Keystone engine 0.9.2 (incorrectly) defaults to radix 16. so we'd better off only using 0x-prefixed integers from now.
# See the related PR: https://github.com/keystone-engine/keystone/pull/382
# and the related issue: https://github.com/keystone-engine/keystone/issues/436


class DataScoutAnalyzer(Analyzer):
    """
    Grabs the environment and auxiliary vector from the target.
    """

    REQUIRED_IMPLANT = "shellphish_qemu"

    def __init__(self, target, analyzer=None):
        super().__init__(target)
        self.env = None
        self.argv = None
        self.auxv = None
        self.map = None
        self.analyzer = analyzer

    def _pushstr(self, s):
        """
        push a string onto stack
        """
        def _cutstr(bits, little=True):
            w = bits // 8  # word size
            byte_order = -1 if little else 1
            n = ["0"] + [s[i:i + w].ljust(w, "\0")[::byte_order].encode('utf-8').hex() for i in range(0, len(s), w)][::-1]
            return n

        if self.target.target_arch == 'x86_64':
            elems = _cutstr(64)
            return "".join("mov rax, 0x%s; push rax; " % word for word in elems)
        elif self.target.target_arch == 'i386':
            elems = _cutstr(32)
            return "".join("mov eax, 0x%s; push eax; " % word for word in elems)
        elif self.target.target_arch in ('mips', 'mipsel'):
            elems = _cutstr(32, little=self.target.target_arch != 'mips')
            return "".join("li $t0, 0x%s; addi $sp, $sp, -4; sw $t0, 0($sp);" % word for word in elems)
        elif self.target.target_arch == 'arm':
            elems = _cutstr(32)
            return "".join(f"movw r0, #0x{word} & 0xffff; movt r0, #0x{word} >> 16; push {{r0}};" for word in elems)
        else:
            raise NotImplementedError()

    def read_file_shellcode(self, filename):
        """
        shellcode to read the content of a file
        """
        if self.target.target_arch == 'x86_64':
            return (
                self._pushstr(filename) +
                "mov rdi, rsp; xor rsi, rsi; xor rdx, rdx; mov rax, 2; syscall;" + # fd = open(path, O_RDONLY, 0)
                "mov r12, rax; sub rsp, 0x1000;" + # alloca 0x1000
                "loop_head:" +
                "xor rax, rax; mov rdi, r12; mov rsi, rsp; mov rdx, 0x1000; syscall;" + # n = read(fd, rsp, 0x1000)
                "mov r13, rax;" + # save n
                "mov rax, 1; mov rdi, 1; mov rsi, rsp; mov rdx, r13; syscall;" + # write(1, rsp, n)
                "test r13, r13; jnz loop_head;" # loop untill we are done with the file
            )
        elif self.target.target_arch == 'i386':
            return (
                self._pushstr(filename) +
                "mov ebx, esp; xor ecx, ecx; xor edx, edx; mov eax, 5; int 0x80;" + # n = open(path, O_RDONLY, 0)
                "mov esi, eax; sub esp, 0x1000;" + # alloca 0x1000, fd = esi
                "loop_head:" +
                "mov eax, 3; mov ebx, esi; mov ecx, esp; mov edx, 0x1000; int 0x80;" + # n = read(fd, rsp, 0x1000)
                "mov edi, eax;"+ # save n
                "mov eax, 4; mov ebx, 1; mov ecx, esp; mov edx, edi; int 0x80;" + # write(1, rsp, n)
                "test edi, edi; jnz loop_head;" # loop untill we are done with the file
            )
        elif self.target.target_arch in ('mips', 'mipsel'):
            return (
                self._pushstr(filename) +
                "move $a0, $sp; xor $a1, $a1, $a1; xor $a2, $a2, $a2; li $v0, 0xfa5; syscall;" +  # n = open(path, O_RDONLY, 0)
                "move $s0, $v0; li $a0, 0x1000; sub $sp, $sp, $a0;" + # alloca 0x1000, fd = $s0
                "loop_head:" +
                "li $v0, 0xfa3; move $a0, $s0; move $a1, $sp; li $a2, 0x1000; syscall;" + # n = read(fd, rsp, 0x1000)
                "move $s1, $v0;" + # save n
                "li $v0, 0xfa4; li $a0, 1; move $a1, $sp; move $a2, $s1; syscall;" + # write(1, rsp, n)
                "bne $s1, 0, loop_head;" # loop untill we are done with the file
            )
        elif self.target.target_arch == 'arm':
            return (
                self._pushstr(filename) +
                "mov r0, sp; eor r1, r1; eor r2, r2; mov r7, #5; svc 0;" +  # n = open(path, O_RDONLY, 0)
                "mov r8, r0; sub sp, sp, 0x1000;" + # alloca 0x1000, fd = $r8
                "loop_head:" +
                "mov r7, #3; mov r0, r8; mov r1, sp; mov r2, 0x1000; svc 0;" + # n = read(fd, rsp, 0x1000)
                "mov r9, r0;" + # save n to r9
                "mov r7, #4; mov r0, 1; mov r1, sp; mov r2, r9; svc 0;" + # write(1, rsp, n)
                "cmp r9, #0; bne loop_head;" # loop untill we are done with the file
            )
        else:
            raise NotImplementedError("Unknown target architecure: \"%s\"!" % self.target.target_arch)

    def echo_shellcode(self, what):
        if self.target.target_arch == 'x86_64':
            return (
                self._pushstr(what) +
                "mov rdi, 1; mov rsi, rsp; mov rdx, %#x; mov rax, 1; syscall;" % len(what) # n = write(1, rsp, 0x1000)
            )
        elif self.target.target_arch == 'i386':
            return (
                self._pushstr(what) +
                "mov ebx, 1; mov ecx, esp; mov edx, %#x; mov eax, 4; int 0x80;" % len(what) # n = write(1, esp, 0x1000)
            )
        elif self.target.target_arch in ('mips', 'mipsel'):
            return (
                self._pushstr(what) +
                "li $a0, 1; move $a1, $sp; li $a2, %#x; li $v0, 0xfa4; syscall;" % len(what)  # n = write(1, sp, 0x1000)
            )
        elif self.target.target_arch == 'arm':
            return (
                self._pushstr(what) +
                "mov r0, #1; mov r1, sp; mov r2, #%#x; mov r7, #4; svc 0;" % len(what)  # n = write(1, sp, 0x1000)
            )
        else:
            raise NotImplementedError()

    def brk_shellcode(self):
        if self.target.target_arch == 'x86_64':
            return "mov rax, 0xc; xor rdi, rdi; syscall; mov rdi, rax; add rdi, 0x1000; mov rax, 0xc; syscall;"
        elif self.target.target_arch == 'i386':
            # n = brk 0
            # brk n + 0x1000
            return "mov eax, 0x2d; xor ebx, ebx; int 0x80; mov ebx, eax; add ebx, 0x1000; mov eax, 0x2d; int 0x80;"
        elif self.target.target_arch in ('mips', 'mipsel'):
            # n = brk 0
            # brk n + 0x1000
            return "xor $a0, $a0, $a0; li $v0, 0xfcd; syscall; add $a0, $v0, 0x1000; li $v0, 0xfcd; syscall;"
        elif self.target.target_arch == 'arm':
            # n = brk 0
            # brk n + 0x1000
            return "eor r0, r0; mov r7, #0x2d; svc 0; add r0, #0x1000; mov r7, #0x2d; svc 0;"
        else:
            raise NotImplementedError()

    def exit_shellcode(self, exit_code=42):
        if self.target.target_arch == 'x86_64':
            return "mov rdi, %#x; mov rax, 0x3c; syscall;" % exit_code # exit(code)
        elif self.target.target_arch == 'i386':
            return "mov ebx, %#x; mov eax, 1; int 0x80;" % exit_code # exit(code)
        elif self.target.target_arch in ('mips', 'mipsel'):
            return "li $a0, %#x; li $v0, 0xfa1; syscall;" % exit_code  # exit(code)
        elif self.target.target_arch == 'arm':
            return "mov r0, #%#x; mov r7, #1; svc 0;" % exit_code  # exit(code)
        else:
            raise NotImplementedError()

    def run_shellcode(self, shellcode, aslr=False, **kwargs):
        exit_code = 42

        # build the args
        if self.analyzer:
            args = self.analyzer._build_command()
        else:
            args = self.target.target_args

        # run command within the shellcode context
        with self.target.shellcode_context(args, asm_code=shellcode+self.exit_shellcode(exit_code=exit_code), aslr=aslr, **kwargs) as p:
            output, stderr = p.communicate()
            if p.returncode != exit_code:
                raise ArchrError("DataScout failed to get info from the target process.\n"
                                 "stdout: %s\nstderr: %s" % (output, stderr))

        return output

    def fire(self, aslr=False, **kwargs): #pylint:disable=arguments-differ
        if self.target.target_os == 'cgc':
            return [], [], b'', {}


        if not self.argv:
            output = self.run_shellcode(self.read_file_shellcode("/proc/self/cmdline"), aslr=aslr, **kwargs)
            self.argv = output.split(b'\0')[:-1]

        if not self.env:
            output = self.run_shellcode(self.read_file_shellcode("/proc/self/environ"), aslr=aslr, **kwargs)
            self.env = output.split(b'\0')[:-1]

        if not self.auxv:
            output = self.run_shellcode(self.read_file_shellcode("/proc/self/auxv"), aslr=aslr, **kwargs)
            self.auxv = output

        if not self.map:
            output = self.run_shellcode(self.brk_shellcode()+self.read_file_shellcode("/proc/self/maps"), aslr=aslr, **kwargs)
            self.map = parse_proc_maps(output)

        return self.argv, self.env, self.auxv, self.map

from ..utils import parse_proc_maps



import subprocess
import struct
import re
import logging
import cle
import io

from . import strace_parser

l = logging.getLogger("archr.utils")


def parse_ldd(mem_map_str):
    entries = [l.strip() for l in mem_map_str.decode('utf-8').splitlines()]
    parsed = { }
    for entry in entries:
        if '=>' in entry:
            libname, paren_addr = entry.split('=>')[1].split()
        else:
            libname, paren_addr = entry.split()
        libaddr = int(paren_addr.strip("()"), 16)
        parsed[libname] = libaddr
    return parsed

def parse_proc_maps(proc_str):
    entries = [ l.strip() for l in proc_str.splitlines() ]
    parsed = { }
    for entry in entries:
        what = entry.split()[-1].decode('utf-8')
        addr_range = entry.split()[0]
        start,end = addr_range.split(b"-")
        if what in parsed:
            continue
        if what.startswith("/"):
            parsed[what] = int(start, 16)
        elif what.startswith("["):
            parsed[what] = int(start, 16)
            parsed[what.rstrip("]")+"-end]"] = int(end, 16)
    return parsed

def lib_dependencies(filepath):
    mem_map_str,_ = subprocess.Popen([ "ldd", filepath ], stdout=subprocess.PIPE).communicate()
    return [ lib for lib in parse_ldd(mem_map_str) if lib != "linux-vdso.so.1" ]

def hook_entry(binary, asm_code=None, bin_code=None):
    main_bin = io.BytesIO(binary)
    b = cle.Loader(main_bin, auto_load_libs=False, perform_relocations=False)
    start_addr = b.main_object.addr_to_offset(b.main_object.entry)
    arch = b.main_object.arch
    if arch.name in ('ARMHF', 'ARMEL') and arch.is_thumb(start_addr): # OMG, thumb mode is a disaster
        start_addr &= (~1) # recover the real address
        main_bin.seek(start_addr)
        padding = (4 - (start_addr + 8) % 4) % 4 # we hardcode the shellcode so that its length is 8

        # we can' use arch.asm here because the shellcode THUMB, 8+padding-4 because the shellcode has length 8+padding,
        # we also need to take into account that in arm, pc points to two instructions ahead, which is 4 bytes in thumb mode
        main_bin.write(b'xF\x00\xf1' + struct.pack('<H', 8+padding-4) + b'\x00G' + b'A'*padding)

        # now place our payload after this mini shellcode
        start_addr += 8 + padding
    main_bin.seek(start_addr)
    main_bin.write(b.main_object.arch.asm(asm_code) if asm_code else bin_code)
    main_bin.seek(0)
    return main_bin.read()

def filter_strace_output(lines):
    """
    a function to filter QEMU logs returning only the strace entries

    Parameters
    ----------
    lines : list
        a list of strings representing the lines from a QEMU log/trace.

    Returns
    -------
    list
        a list of strings representing only the strace log entries
        the entries will also be cleaned up if a page dump occurs in the middle of them
    """

    #we only want the strace lines, so remove/ignore lines that start with the following:
    line_starts= ['^[\d,a-f]{16}-', # pylint: disable=anomalous-backslash-in-string
                      '^[\d,a-f]{8}-', # pylint: disable=anomalous-backslash-in-string
                      '^page',
                      '^start',
                      '^host',
                      '^Locating',
                      '^guest_base',
                      '^end_',
                      '^brk',
                      '^entry',
                      '^argv_',
                      '^env_',
                      '^auxv_',
                      '^Trace',
                      '^--- SIGSEGV',
                      '^qemu'
                      ]
    filter_string = '|'.join(line_starts)

    filtered = []
    prev_line = ""
    for line in lines:
        if re.match(filter_string,line):
            continue
        # workaround for https://gitlab.com/qemu-project/qemu/-/issues/654
        if re.search("page layout changed following target_mmap",line):
            prev_line = line.replace("page layout changed following target_mmap","")
            continue
        # The strace_parser does not properly handle _newselect strace lines (It's format differs significantly).
        # so, workaround by simply removing those lines. We currently do not use them.
        if re.search("newselect", line):
            continue #ignore _newselect syscalls
        if re.match('^ = |^= ', line):
            line = prev_line+line

        filtered.append(line)
    return filtered


def get_file_maps(strace_log_lines):
    """
    a function to return a mapping of filenames and associated mmaped addressed for process under QEMU
    this is accomplished by tracking the filenames through file-desciptors across mmap() calls.

    Parameters
    ----------
    strace_log_lines : list
        a list of strings representing only the strace info logged by QEMU

    Returns
    -------
    dict
        a dictionary of filenames and mmapped addresses associated with each file

    """
    files = {
        'open': {},
        'closed': {},
        'sockets': []
    }

    entries = strace_parser.parse(strace_log_lines)
    entries = [entry for entry in entries if entry.syscall in ('openat','open','socket','mmap','mmap2','close')]

    for entry in entries:
        # for an openat, create a dict entry for the file descriptor
        # the entry should be a tuple of the filename, and mmaps (initially empty)
        if entry.syscall == 'openat':
            fd = entry.syscall.result
            # only care about file descriptors other than STDIN,STDOUT,STDERR
            # also ignore errors
            if fd >= 3:
                #use only the base filename
                filename = entry.syscall.args[1].split("/")[-1]
                #tracking if an executable page was ever mapped from the file descriptor
                files['open'][fd] = [filename,[]]

        elif entry.syscall == 'open':
            fd = entry.syscall.result
            # only care about file descriptors other than STDIN,STDOUT,STDERR
            # also ignore errors
            if fd >= 3:
                #use only the base filename
                filename = entry.syscall.args[0].split("/")[-1]
                #tracking if an executable page was ever mapped from the file descriptor
                files['open'][fd] = [filename,[]]

        elif entry.syscall == 'socket':
            fd = entry.syscall.result
            if fd >= 3:
                files['sockets'].append(fd)

        # if a file descriptor is closed, we need to remove it from the open files dictionary
        # we want to track the mmaps, so move it to 'closed' by file name since the file descriptor can be re-used.
        elif entry.syscall == 'close':
            fd = entry.syscall.args[0]
            # only care about file descriptors other than STDIN,STDOUT,STDERR
            if fd >= 3:
                if fd in files['sockets']:
                    files['sockets'].remove(fd)
                else:
                    filename = files['open'][fd][0]
                    mmaps = files['open'][fd][1]

                    # if we never mapped any pages, then we don't care about it.
                    if mmaps:
                        # otherwise move to 'closed'
                        files['closed'][filename] = mmaps

                    del files['open'][fd]


        # we can use the file descriptor to look up the dict entry to update the mmaps
        #TODO: track sizes which should be capturable from the mmap arguments
        elif entry.syscall in ('mmap', 'mmap2'):
            # only care about valid file descriptors
            fd = entry.syscall.args[4]
            if fd >= 3:
                files['open'][fd][1].append(entry.syscall.result)

    #lets "close" everything that never got closed
    for fd,(filename,mmaps) in files['open'].items():
        files['closed'][filename] = mmaps

    return files['closed']


def hook_addr(binary, addr, asm_code=None, bin_code=b''):
    main_bin = io.BytesIO(binary)
    loader = cle.Loader(main_bin, auto_load_libs=False, perform_relocations=False)
    offset = loader.main_object.addr_to_offset(addr)
    main_bin.seek(offset)
    main_bin.write(loader.main_object.arch.asm(asm_code) if asm_code else bin_code)
    main_bin.seek(0)
    return main_bin.read()



__version__ = "9.2.18"

import logging

_LOG = logging.getLogger("archr")

try:
    import angr
    _angr_available = True
except ImportError:
    _LOG.debug("angr import failed. angr support disabled")
    _angr_available = False

try:
    import qtrace
    _qtrace_available = True
except ImportError:
    _LOG.debug("qtrace import failed. qtrace support disabled")
    _qtrace_available = False

from . import targets
from . import implants
from . import utils
from . import analyzers

# backwards compatibility
arrows = implants
arsenal = analyzers



import contextlib
import subprocess
import tempfile
import shutil
import os

@contextlib.contextmanager
def bundle(implant_name: str):
    bundle_path = tempfile.mkdtemp(prefix="archr_bundle_")

    bundle_base_dir = os.path.dirname(__file__)
    if os.path.exists(os.path.join(bundle_base_dir, implant_name)):
        subprocess.check_call([f"{bundle_base_dir}/{implant_name}/bundle", bundle_path])
    else:
        subprocess.check_call([f"{bundle_base_dir}/{implant_name}_bundle", bundle_path])

    yield bundle_path

    shutil.rmtree(bundle_path)

@contextlib.contextmanager
def bundle_binary(path: str):
    bundle_path = tempfile.mkdtemp(prefix="archr_bundle_binary_")

    bundle_base_dir = os.path.dirname(__file__)
    subprocess.check_call([f"{bundle_base_dir}/GENERIC/bundle", bundle_path, path])
    yield bundle_path

    shutil.rmtree(bundle_path)



import time
import logging
from abc import abstractmethod

import nclib

l = logging.getLogger("archr.target.actions")

class ActionError(BaseException):
    pass

class Action:
    def __init__(self):
        self.interaction = None

    @abstractmethod
    def perform(self):
        """
        Specify how to perform an action
        """
        raise NotImplementedError()

class OpenChannelAction(Action):
    def __init__(self, channel_name=None):
        super().__init__()
        self.channel_name = channel_name

    def _open_channel(self, channel_name):
        if channel_name == 'stdio':
            process = self.interaction.process
            if process is None:
                raise ValueError("Can't get stdio for remote process")
            channel = nclib.merge([process.stdout, process.stderr], sock_send=process.stdin)
        elif ':' in channel_name:
            target = self.interaction.target
            kind, idx = channel_name.split(':', 1)
            if kind in ('tcp', 'tcp6'):
                ipv6 = kind == 'tcp6'
                mapping = target.tcp_ports
                udp = False
            elif kind in ('udp', 'udp6'):
                ipv6 = kind == 'udp6'
                mapping = target.udp_ports
                udp = True
            else:
                raise ValueError("Bad channel", kind)

            address = target.ipv6_address if ipv6 else target.ipv4_address
            # if we run in network_mode=host we don't get an IP
            if not address:
                address = 'localhost'

            try:
                port = mapping[int(idx)]
            except ValueError as e:
                raise ValueError("Channel number is not a number", channel_name) from e
            except LookupError as e:
                raise ValueError("No mapping for channel number", kind, idx) from e

            channel = nclib.Netcat((address, port), udp=udp, ipv6=ipv6, retry=30)
        else:
            raise ValueError("Bad channel", channel_name)

        logger = nclib.logger.StandardLogger(nclib.simplesock.SimpleLogger('archr.log'))
        channel.add_logger(logger)
        return channel

    def perform(self):
        if not self.interaction:
            raise ActionError("No interaction context to perform %s" % self.__class__)
        if self.channel_name is None:
            self.channel_name = self.interaction.default_channel_name
        l.debug("[OpenChannelAction] openning channel: %s", self.channel_name)
        channel = self._open_channel(self.channel_name)
        self.interaction._channels[self.channel_name] = channel
        return channel

class SendAction(Action):

    def __init__(self, data, channel_name=None):
        super().__init__()
        self.channel_name = channel_name
        self.data = data

    def perform(self):
        if not self.interaction:
            raise ActionError("No interaction context to perform %s" % self.__class__)
        if self.channel_name is None:
            self.channel_name = self.interaction.default_channel_name
        l.debug("[SendAction] sending data to channel %s: %s", self.channel_name, self.data)
        channel = self.interaction.get_channel(self.channel_name)
        channel.write(self.data)

class WaitAction(Action):
    def __init__(self, seconds):
        super().__init__()
        self.seconds = seconds

    def perform(self):
        if not self.interaction:
            raise ActionError("No interaction context to perform %s" % self.__class__)
        l.debug("[WaitAction] waiting for %d seconds", self.seconds)
        time.sleep(self.seconds)

class CloseChannelAction(Action):
    def __init__(self, channel_name=None):
        super().__init__()
        self.channel_name = channel_name

    def perform(self):
        if not self.interaction:
            raise ActionError("No interaction context to perform %s" % self.__class__)
        if self.channel_name is None:
            self.channel_name = self.interaction.default_channel_name
        l.debug("[CloseChannelAction] closing channel: %s", self.channel_name)
        channel = self.interaction.get_channel(self.channel_name)
        channel.shutdown_wr()
        channel.close()
        assert channel.closed
        self.interaction._channels.pop(self.channel_name)
        # TODO: if the channel type is stdin, close the process as well.



import subprocess
import logging

from .actions import OpenChannelAction


l = logging.getLogger("archr.target.flight")

class InteractionError(BaseException):
    pass
class Interaction:
    """
    An Interaction specifies how to interact with a target
    An Interaction is generally the result of ContextAnalyzer.fire_context.
    It has a result field, which is normally returned from ContexAnalyzer.fire.
    """
    def __init__(self, target, process, actions=None, result=None):
        self.target = target
        self.process = process
        self._channels = {}
        self.result = result

        if actions is None:
            l.warning("No actions specified, make sure this is what you want!")
            actions = ()
        self.actions = actions
        assert type(actions) in (list, tuple), "actions must be a list or a tuple"
        for act in self.actions:
            act.interaction = self

    def get_channel(self, channel_name):
        # TODO: this doesn't quite work bc we will want to open multiple connections to tcp:0 for example
        # how to represent this...?
        channel = self._channels.get(channel_name, None)
        if channel is not None:
            return channel

        # for backward compatibility where the code interact with the flight object directly
        act = OpenChannelAction(channel_name=channel_name)
        act.interaction = self
        channel = act.perform()
        if channel:
            return channel

        raise InteractionError(f"channel {channel_name} is not open")

    @property
    def default_channel_name(self):
        if self.target.tcp_ports and self.target.ip_version == 4:
            name = 'tcp:0'
        elif self.target.tcp_ports and self.target.ip_version == 6:
            name = 'tcp6:0'
        elif self.target.udp_ports and self.target.ip_version == 4:
            name = 'udp:0'
        elif self.target.udp_ports and self.target.ip_version == 6:
            name = 'udp6:0'
        elif self.process:
            name = 'stdio'
        else:
            raise ValueError("Target has no default channel defined")
        return name

    @property
    def default_channel(self):
        channel_name = self.default_channel_name
        return self.get_channel(channel_name)

    def stop(self, timeout=1, timeout_exception=True):
        for sock in self._channels.values():
            if not sock.closed:
                sock.shutdown_wr()
        if self.process is not None:
            self.process.stdin.close()
            #time.sleep(2)
            #if self.process.poll() is None:
            #    print("Hung process")
            #    import ipdb; ipdb.set_trace()
            try:
                self.process.wait(timeout=timeout)
            except subprocess.TimeoutExpired:
                self.process.terminate()
                self.process.wait()
                if timeout_exception:
                    raise

    def start(self):

        # sanity check
        ret = self.process.poll()
        if ret is not None:
            l.error("The target process crashed with return value: %d", ret)
            stdout, stderr = self.process.communicate()
            l.debug("stdout:\n%s", stdout.decode())
            l.debug("stderr:\n%s", stderr.decode())
            raise ValueError("The target process crashed before communication")

        for act in self.actions:
            act.perform()

# backward compatibility
Flight = Interaction



import os
import time
import logging
from contextlib import contextmanager
from archr.targets.actions import OpenChannelAction, SendAction, CloseChannelAction

l = logging.getLogger(name=__name__)

class Analyzer:
    REQUIRED_IMPLANT = None
    REQUIRED_BINARY = None

    def __init__(self, target, implant_bundle=None, implant_binary=None):
        """
        Initializes the analyzer.
        :param Target target: the target to work on
        """
        self.target = target
        if implant_bundle is not None:
            self.REQUIRED_IMPLANT = implant_bundle
        if implant_binary is not None:
            self.REQUIRED_BINARY = implant_binary
        self.nock()

    def nock(self):
        """
        Prepare the implant (inject it into the target).
        """
        if self.REQUIRED_IMPLANT:
            with implants.bundle(self.REQUIRED_IMPLANT) as b:
                self.target.inject_path(b, os.path.join(self.target.tmpwd, self.REQUIRED_IMPLANT))
        if self.REQUIRED_BINARY:
            with implants.bundle_binary(self.REQUIRED_BINARY) as b:
                self.target.inject_path(b, os.path.join(self.target.tmpwd, os.path.basename(self.REQUIRED_BINARY)))

    def fire(self, *args, **kwargs):
        """
        Fire the analyzer at the target.
        """
        raise NotImplementedError()


class ContextAnalyzer(Analyzer):
    """
    A Analyzer base class for analyzers that implement a fire_context instead of a fire.
    Provides a default .fire() that replays a testcase.
    """

    def fire(self, *args, testcase=None, pre_fire_hook=None, channel=None, delay=0, actions=None, **kwargs): #pylint:disable=arguments-differ
        if actions is None and testcase is not None:
            if type(testcase) is bytes:
                open_act = OpenChannelAction(channel_name=channel)
                send_act = SendAction(testcase, channel_name=channel)
                actions = [open_act, send_act]
            elif type(testcase) is list:
                open_act = OpenChannelAction(channel_name=channel)
                actions = [open_act]
                for write in testcase:
                    actions.append(SendAction(write, channel_name=channel))
            else:
                raise TypeError("Unsupported type for testcase")

        kwargs['actions'] = actions

        with self.fire_context(*args, **kwargs) as flight:
            if delay:
                l.info("sleep for %d seconds waiting for the target to initialize", delay)
                time.sleep(delay)  # wait for the target to initialize
            if pre_fire_hook is not None:
                pre_fire_hook(self, flight, channel=channel, testcase=testcase)
            self._fire_testcase(flight, channel=channel)

        return flight.result

    def _fire_testcase(self, flight, channel=None): #pylint:disable=no-self-use
        flight.start()
        return flight.get_channel(channel)

    @contextmanager
    def fire_context(self, *args, **kwargs):  # -> ContextManager[Flight]:
        """
        A context manager for the analyzer. Should yield a Flight object.
        """
        with self.target.flight_context(*args, **kwargs) as flight:
            yield flight


from .. import _angr_available, _qtrace_available

if _angr_available:
    from .angr_project import angrProjectAnalyzer
    from .angr_state import angrStateAnalyzer
    from .angr_ultimate_tracer import angrUltimateTracerAnalyzer
if _qtrace_available:
    from .qtrace import QTraceAnalyzer
from .qemu_tracer import QEMUTracerAnalyzer
from .bintrace_qemu_tracer import BintraceQEMUTracerAnalyzer
from .datascout import DataScoutAnalyzer
from .gdbserver import GDBServerAnalyzer
from .core import CoreAnalyzer
from .ltrace import LTraceAnalyzer, LTraceAttachAnalyzer
from .strace import STraceAnalyzer, STraceAttachAnalyzer
from .input_fd import InputFDAnalyzer
from .rr import RRTracerAnalyzer, RRReplayAnalyzer
from .gdb import GDBAnalyzer
from .tcpdump import TCPDumpAnalyzer
from .udp_tcp_convert import UDPTCPConvert
from .. import implants

# backwards compatibility
if _angr_available:
    angrProjectBow = angrProjectAnalyzer
    angrStateBow = angrStateAnalyzer
    angrUltimateTracerBow = angrUltimateTracerAnalyzer
QEMUTracerBow = QEMUTracerAnalyzer
BintraceQEMUTracerBow = BintraceQEMUTracerAnalyzer
DataScoutBow = DataScoutAnalyzer
GDBServerBow = GDBServerAnalyzer
CoreBow = CoreAnalyzer
UDPTCPConvertBow = UDPTCPConvert
LTraceBow = LTraceAnalyzer
LTraceAttachBow = LTraceAttachAnalyzer
STraceBow = STraceAnalyzer
STraceAttachBow = STraceAttachAnalyzer
InputFDBow = InputFDAnalyzer
RRTracerBow = RRTracerAnalyzer
RRReplayBow = RRReplayAnalyzer
GDBBow = GDBAnalyzer
